/*#######################################################################################
AVR DCF77 Clock 

Copyright (C) 2005 Ulrich Radig
Modified for dcf77usb by Armin Diehl (diehl@nordrhein.de)

#######################################################################################*/

#include "main.h"
#include "clock.h"
#include <inttypes.h>
#include "uart.h"

//Die Uhrzeit seht in folgenden Variablen
unsigned char ss	= 0;	//global var for seconds
unsigned char mm	= 0;	//               minutes
unsigned char hh	= 0;	//               hours
unsigned char day	= 0;	//               day
unsigned char mon	= 0;	//               month
unsigned int year	= 0;	//               year
unsigned char changeMEST_MET = 0;
unsigned char MEST           = 0;
unsigned char MET            = 0;
unsigned char leapSecond     = 0;
unsigned char weekday        = 0;

//Bitzhler fr RX Bit
unsigned char rx_bit_counter = 0;

//64 Bit fr DCF77 bentigt werden 59 Bits
unsigned long long dcf_rx_buffer = 0;

//Hilfs Sekunden Counter
unsigned int h_ss = 0;

unsigned char rx_ok = 0;               // needed by main
unsigned char ss_counter = 0;
unsigned char time_was_set = 0;

uint8_t  ms10counter = 0;              // used to check signal length
uint8_t  ms10counter2 = 0;
uint8_t  SecsFromLastSignal = 0;
uint8_t  newDCFData = 0;

//############################################################################
//Overflow Interrupt, called every 10 ms for counters
SIGNAL (SIG_OVERFLOW0) {
//############################################################################
  TCNT0 = 256 - (SYSCLK / 1024 / 100);  // about 10ms
  if (ms10counter < 255) ms10counter++;
  if (ms10counter2 < 255) ms10counter2++;
  //if (PORTB & (1<<5)) PORTB &= ~(1<<5); else PORTB |= (1<<5);  /* was test on stk500 */
}

void setReceiveError(void) {
  rx_ok = 0;
  PORTD &= ~(1<<7);
}

void setReceiveOk(void) {
  rx_ok = 1;
  PORTD |= (1<<7);
}


void parseDCFData (void) {
  struct  DCF77_Bits *rx_buffer;
  rx_buffer = (struct DCF77_Bits *)(unsigned long long)&dcf_rx_buffer;
		
  if (flags.parity_P1 == rx_buffer->P1 &&
	flags.parity_P2 == rx_buffer->P2 &&
	flags.parity_P3 == rx_buffer->P3)
		//Alle 59Bits empfangen stellen der Uhr nach DCF77 Buffer
  {
    //Berechnung der Minuten BCD to HEX
    mm = rx_buffer->Min-((rx_buffer->Min/16)*6);
    //Berechnung der Stunden BCD to HEX
    hh = rx_buffer->Hour-((rx_buffer->Hour/16)*6);
    //Berechnung des Tages BCD to HEX
    day= rx_buffer->Day-((rx_buffer->Day/16)*6);
    //Berechnung des Monats BCD to HEX
    mon= rx_buffer->Month-((rx_buffer->Month/16)*6);
    //Berechnung des Jahres BCD to HEX
    year= 2000 + rx_buffer->Year-((rx_buffer->Year/16)*6);
    //Sekunden werden auf 0 zurckgesetzt
    ss = 0;
    changeMEST_MET = rx_buffer->A1;
    MEST = rx_buffer->Z1;
    MET = rx_buffer->Z2;
    leapSecond = rx_buffer->A2;
    weekday = rx_buffer->Weekday-((rx_buffer->Weekday/16)*6);
    setReceiveOk();
    newDCFData = 1;
  } else {
    setReceiveError();
  }
    
   //zurcksetzen des RX Bit Counters
   rx_bit_counter = 0;
   //Lschen des Rx Buffers
   dcf_rx_buffer = 0;
   
}

//############################################################################
//Overflow Interrupt will be called every second
//oder jede Sekunde
SIGNAL (SIG_OVERFLOW1)
//############################################################################
{	
		
  //Reset Timer
  TCNT1 = 65535 - (SYSCLK / 1024);
  if (! newDCFData) {
    Add_one_Second();
    SecsFromLastSignal++;
    if (SecsFromLastSignal > 61) setReceiveError();
  } else {
    newDCFData = 0;
    SecsFromLastSignal = 0;
  }
  ss_counter++;  // Main.c will send time
}

//############################################################################
//DCF77 Modul empfngt Trger 
SIGNAL (DCF77_INT)
//############################################################################
{
	//Auswertung der Pulseweite 
	if (INT0_CONTROL == INT0_RISING_EDGE)
	{
		ms10counter = 0;                           // to detect signal length
		if (rx_bit_counter == 59) {
		  parseDCFData ();                         // convert and reset Data
		  TCNT1 = 0xfffe;                          // sync 16 Bit Timer interrupt
		  sei();
		}
		PORTD &= ~(1<<6);                          // Show Signal
		//Naechster Interrupt wird ausgelst bei abfallender Flanke
		INT0_CONTROL = INT0_FALLING_EDGE;
		// Reset Buffer if time between two signals is less than 0,5 seconds or
		// longer than 1,5 seconds
		if ((ms10counter2 > 150) || (ms10counter2 < 50)) {
		   //zurcksetzen des RX Bit Counters
	           rx_bit_counter = 0;
	           //Lschen des Rx Buffers
	           dcf_rx_buffer = 0;
	           //setReceiveError();
		}
		ms10counter2 = 0;                          // to detect time between 2 signals
	}
	else
	{
		//Auslesen der Pulsweite von ansteigender Flanke zu abfallender Flanke
		//unsigned int pulse_wide = TCNT1;
		unsigned int pulse_wide = ms10counter;
		PORTD |= (1<<6);                          // show signal
		//Parity speichern
		//beginn von Bereich P1/P2/P3
		if (rx_bit_counter ==  21 || rx_bit_counter ==  29 || rx_bit_counter ==  36) {
		  flags.parity_err = 0;
		}
		//Speichern von P1
		if (rx_bit_counter ==  28) {flags.parity_P1 = flags.parity_err;};
		//Speichern von P2
		if (rx_bit_counter ==  35) {flags.parity_P2 = flags.parity_err;};
		//Speichern von P3
		if (rx_bit_counter ==  58) {flags.parity_P3 = flags.parity_err;};
		//Check if 0 or 1 received
		//0 = 100ms
		//1 = 200ms
		//> 150ms will be counted as 1
		if (pulse_wide > 15) {
			//Schreiben einer 1 im dcf_rx_buffer an der Bitstelle rx_bit_counter
			dcf_rx_buffer = dcf_rx_buffer | ((unsigned long long) 1 << rx_bit_counter);
			//Toggele Hilfs Parity
			flags.parity_err = flags.parity_err ^ 1;
		}
		//Naechster Interrupt wird ausgeloest bei ansteigender Flanke
		INT0_CONTROL = INT0_RISING_EDGE;
		//RX Bit Counter wird um 1 incrementiert
		rx_bit_counter++;
		
	}
};

//############################################################################
//add one second (used if invalid dcf signal received)
void Add_one_Second (void)
//############################################################################
{
	ss++;//Addiere +1 zu Sekunden
	if (ss == 60)
	{
		ss = 0;
		mm++;//Addiere +1 zu Minuten
		if (mm == 60)
		{
			mm = 0;
			hh++;//Addiere +1 zu Stunden
			if (hh == 24)
			{
				hh = 0;
			}
		}
	}
	ss_counter++;  // main.c will send time
};

//############################################################################
//Initialize second and 10 ms counter
void Start_Clock (void)
//############################################################################
{
	//Interrupt DCF77 einschalten auf ansteigende Flanke
	DCF77_INT_ENABLE();
	INT0_CONTROL = INT0_RISING_EDGE;
		
	//Interrupt Overfolw enable for Timer0 and Timer1
	timer_enable_int(_BV(TOIE1) | _BV(TOIE0));
	// Prescaler = 1024 
	TCCR1B |= (1<<CS10 | 0<<CS11 | 1<<CS12);
	TCCR0  |= (1<<CS00 | 0<<CS01 | 1<<CS02);
	//Timer1 every second (precise @ 16MHz or 4,194304 MHz)
	TCNT1 = 65535 - (SYSCLK / 1024);
	TCNT0 = 256 - (SYSCLK / 1024 / 100);  // about 10ms
	PORTB &= ~(1<<2);                     // show that no signal was received
	return;
};


